/*
 * ts7200_flash.c - mapping for TS-7200 SBCs (8mb NOR flash)
 * No platform_device resource is used here. All is hardcoded.
 *
 * (c) Copyright 2006  Matthieu Crapet <mcrapet@gmail.com>
 * Based on ts5500_flash.c by Sean Young <sean@mess.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <asm/io.h>
#include <asm/sizes.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/partitions.h>

#define WINDOW_ADDR          0x60000000
#define WINDOW_SIZE          SZ_8M
#define WINDOW_READABLE_SIZE (WINDOW_SIZE/SZ_1M)


static struct mtd_info *mymtd;

static struct map_info ts7200nor_map = {
  .name = "Full TS-7200 NOR flash",
  .size = WINDOW_SIZE,
  .bankwidth = 2,
  .phys = WINDOW_ADDR,
};

/*
 * MTD partitioning stuff
 */
#ifdef CONFIG_MTD_PARTITIONS
static struct mtd_partition static_partitions[] =
{
  {
    .name   = "TS-BOOTROM",
    .offset = 0,
    .size   = 0x20000,
    .mask_flags = MTD_WRITEABLE,  /* force read-only */
  },
  {
    .name   = "RootFS",
    .offset = 0x20000,
    .size   = 0x600000,
  },
  {
    .name   = "Redboot",
    .offset = 0x620000,
    .size   = MTDPART_SIZ_FULL,   /* up to the end */
  },
};
#endif

int __init init_ts7200nor(void)
{
  printk(KERN_NOTICE "TS-7200 flash mapping: %dmo at 0x%x\n", WINDOW_READABLE_SIZE, WINDOW_ADDR);

  ts7200nor_map.virt = ioremap(WINDOW_ADDR, WINDOW_SIZE);
  if (!ts7200nor_map.virt) {
    printk("ts7200_flash: failed to ioremap\n");
    return -EIO;
  }

  simple_map_init(&ts7200nor_map);
  mymtd = do_map_probe("cfi_probe", &ts7200nor_map);
  if (mymtd) {
    mymtd->owner = THIS_MODULE;
    add_mtd_device(mymtd);
#ifdef CONFIG_MTD_PARTITIONS
    return add_mtd_partitions(mymtd, static_partitions, ARRAY_SIZE(static_partitions));
#else
    return 0;
#endif
  }

  iounmap((void *)ts7200nor_map.virt);
  return -ENXIO;
}

static void __exit cleanup_ts7200nor(void)
{
  if (mymtd) {
    del_mtd_device(mymtd);
    map_destroy(mymtd);
  }
  if (ts7200nor_map.virt) {
    iounmap((void *)ts7200nor_map.virt);
    ts7200nor_map.virt = 0;
  }
}

module_init(init_ts7200nor);
module_exit(cleanup_ts7200nor);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Matthieu Crapet <mcrapet@gmail.com>");
MODULE_DESCRIPTION("MTD map driver for TS-7200 board (" __MODULE_STRING(WINDOW_READABLE_SIZE) "MB flash version)");
