/*
 *  EP93xx SPI driver
 *
 * (c) Copyright 2008  Matthieu Crapet <mcrapet@gmail.com>
 * Based on pxa2xx_spi.c by Stephen Street / StreetFire Sound Labs
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Notes:
 *  - Uses SSP IP of processor
 *  - Restricted to SPI master mode
 *  - No DMA transfer
 *  - No power management support
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/device.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/platform_device.h>
#include <linux/spi/spi.h>
#include <linux/clk.h>
#include <asm/irq.h>
#include <mach/hardware.h>
#include <mach/spi.h>

#include <linux/sched.h>
#include <linux/spinlock.h>
#include <linux/workqueue.h>

#include "spi_ep93xx.h"


struct ep93xx_spi {
  struct spi_master *master;    /* SPI framework hookup */
  void __iomem *ioaddr;         /* Virtual base address to SSP registers */
  u32 freq_max;
  u32 freq_min;
  struct clk *clk;

  struct workqueue_struct *workqueue;
  struct work_struct work;
  spinlock_t lock;
  struct list_head queue;

  struct ep93xx_spi_chip *cs_chip; /* Chip Select function */
};

static inline u16 read_reg(void __iomem *base, off_t offset)
{
	return __raw_readw(base + offset);
}

static inline void write_reg(u16 v, void __iomem *base, off_t offset)
{
	__raw_writew(v, base + offset);
}

/*
 *  compute SCR and CPSDVR bits to setup spi clock based on main input clock rate
 *  that was specified in platform data structure
 *  according to datasheet:
 *    tempclk = sspclk / cpsdvr
 *    spiclk = tempclk / (scr + 1)
 *    SCR valid range is 0..255
 *    CPSDVR valid range is 2..254
 */
static int spi_speed_set(struct ep93xx_spi *drv_data, unsigned speed_hz)
{
	unsigned long mainclk_hz = clk_get_rate(drv_data->clk);
	u32 cpsdvr, scr;
	u16 ssp_cr0;

	for (cpsdvr = 2; cpsdvr <= 254; cpsdvr+=2) {
		scr = DIV_ROUND_UP(mainclk_hz / speed_hz, cpsdvr);
		/* now we have SCR+1 in scr, so count with that */
		if (scr == 0) { 	/* speed_hz too big */
			return -EINVAL;
		}
		if (scr <= (255 + 1))
			break;		/* we have valid scr and cpsdvr */
	}
	if (cpsdvr > 254) {
		/* speed_hz is too small, set to minimum speed */
		scr = 256;
		cpsdvr = 254;
	}
	scr--;
	write_reg(cpsdvr, drv_data->ioaddr, SSPCPSR);
	ssp_cr0 = read_reg(drv_data->ioaddr, SSPCR0);
	ssp_cr0 &= ~(SSP_CONTROL_SCR(0xff));
	write_reg((ssp_cr0 | SSP_CONTROL_SCR(scr)), drv_data->ioaddr, SSPCR0);

	return 0;
}

static irqreturn_t ssp_int(int irq, void *dev_id)
{
	struct ep93xx_spi *drv_data = dev_id;
	write_reg(SSP_SSPIxx_RORIS, drv_data->ioaddr, SSPIxR); /* clear it */

	printk(KERN_WARNING "SSP overrun\n");
	return IRQ_HANDLED;
}

static int transfer_one_work(struct ep93xx_spi *drv_data, struct spi_message *msg)
{
	struct spi_device *spi = msg->spi;
	struct spi_transfer *xfer;
	int i;
	u8 *p;

	drv_data->cs_chip->cs_control(SPI_CS_ASSERT);

	list_for_each_entry(xfer, &msg->transfers, transfer_list) {
		if (!(xfer->tx_buf || xfer->rx_buf)) {
			dev_dbg(&spi->dev, "missing rx or tx buf\n");
			drv_data->cs_chip->cs_control(SPI_CS_DEASSERT);
			return -EINVAL;
		}

		if (xfer->bits_per_word) {
			u16 v = read_reg(drv_data->ioaddr, SSPCR0);
			v = v & SSP_CONTROL_DSS_MASK;
			v = v | ((xfer->bits_per_word - 1) & SSP_CONTROL_DSS_MASK);
			write_reg(v, drv_data->ioaddr, SSPCR0);
		}

		if (xfer->speed_hz) {
			if (spi_speed_set(drv_data,xfer->speed_hz) != 0){
				dev_err(&spi->dev, "xfer speed hz invalid\n");
				return -EINVAL;
			}
		}

		if (xfer->tx_buf) {
			p = (u8 *)xfer->tx_buf;

			if ((spi->bits_per_word == 16 && xfer->bits_per_word == 0) ||
			    (xfer->bits_per_word == 16)) {
				for (i = 0; i < xfer->len; i+=2)
					write_reg((p[i] << 8) + p[i+1], drv_data->ioaddr, SSPDR);
			} else {
				for (i = 0; i < xfer->len; i++)
					write_reg(p[i], drv_data->ioaddr, SSPDR);
			}
		}

		if (xfer->rx_buf) {
			u16 v;
			p = xfer->rx_buf;

			if ((spi->bits_per_word == 16 && xfer->bits_per_word == 0) ||
			    (xfer->bits_per_word == 16)) {
					for (i = 0; i < xfer->len; i+=2) {
						  v = read_reg(drv_data->ioaddr, SSPDR);
						  p[i] = v >> 8;
						  p[i+1] = v & 0xFF;
					}
			} else {
					for (i = 0; i < xfer->len; i++)
						  p[i] = read_reg(drv_data->ioaddr, SSPDR);
			}
		}

		/* restore device bits_per_word */
		if (xfer->bits_per_word) {
			u16 v = read_reg(drv_data->ioaddr, SSPCR0);
			v = v & SSP_CONTROL_DSS_MASK;
			v |= spi->bits_per_word - 1;
			write_reg(v, drv_data->ioaddr, SSPCR0);
		}

		/* restore device speed_hz */
		if (xfer->speed_hz) {
			if (spi_speed_set(drv_data,spi->max_speed_hz) != 0)
				return -EINVAL;
		}

		dev_dbg(&spi->dev, "transfer: len=%u, tx_buf=%p, rx_buf=%p\n", xfer->len, xfer->tx_buf, xfer->rx_buf);
	}

	if (xfer->delay_usecs)
		udelay(xfer->delay_usecs);
	drv_data->cs_chip->cs_control(SPI_CS_DEASSERT);

	msg->actual_length = 0;
	msg->status = 0;

	if (msg->complete)
		msg->complete(msg->context);

	return 0;
}


static void ssp_work(struct work_struct *work)
{
	struct ep93xx_spi *drv_data = container_of(work, struct ep93xx_spi, work);
	unsigned long flags;

	spin_lock_irqsave(&drv_data->lock, flags);
	while (!list_empty(&drv_data->queue)) {
		struct spi_message *m;

		m = container_of(drv_data->queue.next, struct spi_message, queue);
		list_del_init(&m->queue);
		spin_unlock_irqrestore(&drv_data->lock, flags);

		transfer_one_work(drv_data, m);

		spin_lock_irqsave(&drv_data->lock, flags);
	}
	spin_unlock_irqrestore(&drv_data->lock, flags);
}


static int ssp_transfer(struct spi_device *spi, struct spi_message *m)
{
	struct spi_master *master = spi->master;
	struct ep93xx_spi *drv_data = spi_master_get_devdata(master);
	struct spi_transfer *t;
	unsigned long flags;

	m->actual_length = 0;

	/* check each transfer's parameters */
	list_for_each_entry (t, &m->transfers, transfer_list) {
		u32 speed_hz = t->speed_hz ? t->speed_hz : spi->max_speed_hz;
		u8 bits_per_word = t->bits_per_word ? t->bits_per_word : spi->bits_per_word;

		if (!t->tx_buf && !t->rx_buf && t->len)
			return -EINVAL;
		if (bits_per_word < 4 || bits_per_word > 16)
			return -EINVAL;
		/*if (t->len & ((bits_per_word >> 3) - 1))
			return -EINVAL;*/
		if (speed_hz < drv_data->freq_min || speed_hz > drv_data->freq_max)
			return -EINVAL;
	}

	spin_lock_irqsave(&drv_data->lock, flags);
	list_add_tail(&m->queue, &drv_data->queue);
	queue_work(drv_data->workqueue, &drv_data->work);
	spin_unlock_irqrestore(&drv_data->lock, flags);

	return 0;

}

/* the spi->mode bits understood by this driver: */
#define MODEBITS (SPI_CPOL | SPI_CPHA)

static int ssp_setup(struct spi_device *spi)
{
	struct ep93xx_spi *drv_data = spi_master_get_devdata(spi->master);
	struct ep93xx_spi_chip *chip_info;
	u16 v;

	/* Get controller data */
	chip_info = spi->controller_data;
	if (!chip_info) {
		dev_err(&spi->dev, "setup: controller data required\n");
		return -EINVAL;
	}
	drv_data->cs_chip = chip_info;
	drv_data->cs_chip->cs_control(SPI_CS_INIT);

	if (!spi->bits_per_word) {
		spi->bits_per_word = 8;
	}

	if (spi->bits_per_word < 4 || spi->bits_per_word > 16) {
		  dev_dbg(&spi->dev, "setup: unsupported %d bit words\n",
		  spi->bits_per_word);
		  return -EINVAL;
	}

	if (spi->chip_select > spi->master->num_chipselect) {
		  dev_dbg(&spi->dev, "setup: invalid chipselect %u (%u defined)\n",
		  spi->chip_select,  spi->master->num_chipselect);
		  return -EINVAL;
	}

	if (spi->mode & ~MODEBITS) {
		 dev_dbg(&spi->dev, "setup: unsupported mode bits %x\n",
		 spi->mode & ~MODEBITS);
		 return -EINVAL;
	}

	v = read_reg(drv_data->ioaddr, SSPCR0);

	if (spi->mode & SPI_CPOL)
		v |= SSP_CONTROL_SPO;
	else
		v &= ~SSP_CONTROL_SPO;

	if (spi->mode & SPI_CPHA)
		v |= SSP_CONTROL_SPH;
	else
		v &= ~SSP_CONTROL_SPH;

	v = v & SSP_CONTROL_DSS_MASK;
	v |= spi->bits_per_word - 1;

	write_reg(v, drv_data->ioaddr, SSPCR0);

	if (!spi->max_speed_hz) {
		spi->max_speed_hz = drv_data->freq_min;
	} else if (spi->max_speed_hz > drv_data->freq_max ||
			spi->max_speed_hz < drv_data->freq_min){
		return -EINVAL;
	}

	if (spi_speed_set(drv_data,spi->max_speed_hz) != 0){
		dev_dbg(&spi->dev, "setup: unsupported speed %u\n", spi->max_speed_hz);
		return -EINVAL;
	}

	return 0;
}


static void ssp_cleanup(struct spi_device *spi)
{
	struct ep93xx_spi *drv_data = spi_master_get_devdata(spi->master);
	drv_data->cs_chip = NULL;
}


static int __init spi_ep93xx_probe(struct platform_device *pdev)
{
	struct device *dev = &pdev->dev;
	struct spi_master *master;
	struct ep93xx_spi_data *spi_data = pdev->dev.platform_data;
	struct ep93xx_spi *drv_data = NULL;
	struct resource *memory_resource;
	int irq, status = 0, min_div = 2, max_div = 254*(255+1);

	/* Check I2SonSSP bit (ssp pins and i2s pins are multiplexed)
	   We could force with ep93xx_devcfg_clear_bits */
	if (readl(EP93XX_SYSCON_DEVCFG) & EP93XX_SYSCON_DEVCFG_I2SONSSP)
		return -ENODEV;

	/* Allocate master with space for drv_data */
	master = spi_alloc_master(dev, sizeof(struct ep93xx_spi));
	if (!master) {
		dev_err(&pdev->dev, "can not alloc spi_master\n");
		return -ENOMEM;
	}

	/* Setup register addresses */
	memory_resource = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!memory_resource) {
		dev_err(&pdev->dev, "memory resources not defined\n");
		status = -ENODEV;
		goto out_error_master_alloc;
	}

	drv_data = spi_master_get_devdata(master);
	drv_data->master = master;
	drv_data->ioaddr = (void *)memory_resource->start;
	drv_data->clk = clk_get(&pdev->dev, "sspclk");
	drv_data->freq_max = clk_get_rate(drv_data->clk) / min_div;
	drv_data->freq_min = clk_get_rate(drv_data->clk) / max_div + 1;

	INIT_WORK(&drv_data->work, ssp_work);
	spin_lock_init(&drv_data->lock);
	INIT_LIST_HEAD(&drv_data->queue);

	drv_data->workqueue = create_singlethread_workqueue(master->dev.parent->bus_id);
	if (!drv_data->workqueue){
 		status = -EBUSY;
		goto out_error_master_alloc;
	}

	master->bus_num = pdev->id;
	master->num_chipselect = spi_data->chip_select_num;
	master->cleanup = ssp_cleanup;
	master->setup = ssp_setup;
	master->transfer = ssp_transfer;

	/* Attach to IRQ */
	irq = platform_get_irq(pdev, 0);
	if (irq < 0) {
		dev_err(&pdev->dev, "irq resource not defined\n");
		status = -ENODEV;
		goto out_error_master_alloc;
	}

	status = request_irq(irq, ssp_int, 0, dev->bus_id, drv_data);
	if (status < 0) {
		dev_err(&pdev->dev, "can not get IRQ\n");
		goto out_error_master_alloc;
	}

	/* Load default SSP configuration */
	write_reg(SSP_CONTROL_SSE, drv_data->ioaddr, SSPCR1);
	write_reg(SPI_DEFAULT0, drv_data->ioaddr, SSPCR0);
	write_reg(SPI_DEFAULT_DIVISOR, drv_data->ioaddr, SSPCPSR);
	write_reg(0x00, drv_data->ioaddr, SSPCR1);

	/* Register with the SPI framework */
	platform_set_drvdata(pdev, drv_data);
	status = spi_register_master(master);
	if (status != 0) {
		dev_err(&pdev->dev, "problem registering spi master\n");
		goto out_error_irq_alloc;
	}

	write_reg(SPI_DEFAULT1, drv_data->ioaddr, SSPCR1);
	return status;

out_error_irq_alloc:
	free_irq(irq, drv_data);

out_error_master_alloc:
	spi_master_put(master);
	return status;
}

static int spi_ep93xx_remove(struct platform_device *pdev)
{
	struct ep93xx_spi *drv_data = platform_get_drvdata(pdev);
	int irq;

	if (!drv_data)
		return 0;

	/* Disable SSP (clear SSE bit) */
	write_reg(0x00, drv_data->ioaddr, SSPCR1);

	/* Release IRQ */
	irq = platform_get_irq(pdev, 0);

	if (irq >= 0)
		free_irq(irq, drv_data);

	/* Disconnect from the SPI framework */
	spi_unregister_master(drv_data->master);

	/* Remove the workqueue */
	destroy_workqueue(drv_data->workqueue);

	/* Prevent double remove */
	platform_set_drvdata(pdev, NULL);

	spi_master_put(drv_data->master);

	return 0;
}

static void spi_ep93xx_shutdown(struct platform_device *pdev)
{
	int status = 0;

	if ((status = spi_ep93xx_remove(pdev)) != 0)
		dev_err(&pdev->dev, "shutdown failed with %d\n", status);
}

static struct platform_driver ep93xx_spi_platform_driver = {
	.driver = {
		.name = "ep93xx-spi",
		.bus = &platform_bus_type,
		.owner = THIS_MODULE,
	},
	.remove = __exit_p(spi_ep93xx_remove),
	.shutdown = spi_ep93xx_shutdown,
};

static int __init spi_ep93xx_init(void)
{
	return platform_driver_probe(&ep93xx_spi_platform_driver, spi_ep93xx_probe);
}

static void __exit spi_ep93xx_exit(void)
{
	platform_driver_unregister(&ep93xx_spi_platform_driver);
}

module_init(spi_ep93xx_init);
module_exit(spi_ep93xx_exit);

MODULE_AUTHOR("Matthieu Crapet <mcrapet@gmail.com>");
MODULE_DESCRIPTION("EP93xx SPI Controller Driver");
MODULE_LICENSE("GPL");
MODULE_VERSION("0.23");
